@@ -20,6 +20,7 @@ import json
 import logging
 import os
 import re
+import signal
 import sys
 import tempfile
 import time
@@ -36,7 +37,7 @@ else:
     xrange = range
 
 # Wrapper version
-VERSION = 1
+VERSION = 3
 
 LOG_LEVEL = logging.DEBUG
 STATE_DIR = '/tmp'
@@ -57,8 +58,15 @@ PREALLOCATED_STORAGE_TYPES = (
     )
 
 # Tweaks
-VDSM = False
-DIRECT_BACKEND = not VDSM
+VDSM = True
+# We cannot use the libvirt backend in virt-v2v and have to use direct backend
+# for several reasons:
+# - it is necessary on oVirt host when running as root; and we need to run as
+#   root when using export domain as target (we use vdsm user for other
+#   targets)
+# - SSH transport method cannot be used with libvirt because it does not pass
+#   SSH_AUTH_SOCK env. variable to the QEMU process
+DIRECT_BACKEND = True
 
 
 def error(msg):
@@ -75,19 +83,31 @@ def error(msg):
     sys.exit(1)
 
 
-def make_vdsm():
+def make_vdsm(data):
     """Makes sure the process runs as vdsm user"""
     uid = os.geteuid()
     if uid == VDSM_UID:
-        logging.debug('Already running as vdsm user')
+        # logging.debug('Already running as vdsm user')
         return
     elif uid == 0:
-        logging.debug('Restarting as vdsm user')
-        os.chdir('/')
+        # We need to drop privileges and become vdsm user, but we also need the
+        # proper environment for the user which is tricky to get. The best
+        # thing we can do is spawn another instance. Unfortunately we have
+        # already read the data from stdin.
+        # logging.debug('Starting instance as vdsm user')
         cmd = '/usr/bin/sudo'
         args = [cmd, '-u', 'vdsm']
         args.extend(sys.argv)
-        os.execv(cmd, args)
+        wrapper = subprocess.Popen(args,
+                                   stdin=subprocess.PIPE,
+                                   stdout=subprocess.PIPE,
+                                   stderr=subprocess.PIPE)
+        out, err = wrapper.communicate(json.dumps(data))
+        # logging.debug('vdsm instance finished')
+        sys.stdout.write(out)
+        sys.stderr.write(err)
+        # logging.debug('Terminating root instance')
+        sys.exit(wrapper.returncode)
     sys.stderr.write('Need to run as vdsm user or root!\n')
     sys.exit(1)
 
@@ -123,6 +143,11 @@ class OutputParser(object):
     DISK_PROGRESS_RE = re.compile(br'\s+\((\d+\.\d+)/100%\)')
     NBDKIT_DISK_PATH_RE = re.compile(
         br'nbdkit: debug: Opening file (.*) \(.*\)')
+    OVERLAY_SOURCE_RE = re.compile(
+        br' *overlay source qemu URI: json:.*"file\.path": ?"([^"]+)"')
+    VMDK_PATH_RE = re.compile(
+            br'/vmfs/volumes/(?P<store>[^/]*)/(?P<vm>[^/]*)/'
+            + br'(?P<disk>.*)(-flat)?\.vmdk')
 
     def __init__(self, v2v_log):
         self._log = open(v2v_log, 'rbU')
@@ -150,6 +175,7 @@ class OutputParser(object):
                 except ValueError:
                     logging.exception('Conversion error')
 
+            # VDDK
             m = self.NBDKIT_DISK_PATH_RE.match(line)
             if m is not None:
                 self._current_path = m.group(1).decode()
@@ -157,6 +183,17 @@ class OutputParser(object):
                     logging.info('Copying path: %s', self._current_path)
                     self._locate_disk(state)
 
+            # SSH
+            m = self.OVERLAY_SOURCE_RE.match(line)
+            if m is not None:
+                path = m.group(1).decode()
+                # Transform path to be raltive to storage
+                self._current_path = self.VMDK_PATH_RE.sub(
+                    br'[\g<store>] \g<vm>/\g<disk>', path)
+                if self._current_disk is not None:
+                    logging.info('Copying path: %s', self._current_path)
+                    self._locate_disk(state)
+
             m = self.DISK_PROGRESS_RE.match(line)
             if m is not None:
                 if self._current_path is not None and \
@@ -221,12 +258,14 @@ def sdk_connection(data):
     url = urlparse(data['rhv_url'])
     username = url.username if url.username is not None else 'admin@internal'
     try:
+        insecure = data.get('insecure_connection', False)
         connection = sdk.Connection(
             url=str(data['rhv_url']),
             username=str(username),
             password=str(data['rhv_password']),
             ca_file=str(data['rhv_cafile']),
             log=logging.getLogger(),
+            insecure=insecure,
         )
         yield connection
     finally:
@@ -285,21 +324,27 @@ def write_state(state):
         json.dump(state, f)
 
 
-def wrapper(data, state_file, v2v_log):
+def wrapper(data, state_file, v2v_log, agent_sock=None):
     v2v_args = [
         '/usr/bin/virt-v2v', '-v', '-x',
         data['vm_name'],
-        '-ic', data['vmware_uri'],
-        '--password-file', data['vmware_password_file'],
         '-of', data['output_format'],
         '--bridge', 'ovirtmgmt',
     ]
 
     if data['transport_method'] == 'vddk':
         v2v_args.extend([
+            '-i', 'libvirt',
+            '-ic', data['vmware_uri'],
             '-it', 'vddk',
             '-io', 'vddk-libdir=%s' % '/opt/vmware-vix-disklib-distrib',
             '-io', 'vddk-thumbprint=%s' % data['vmware_fingerprint'],
+            '--password-file', data['vmware_password_file'],
+            ])
+    elif data['transport_method'] == 'ssh':
+        v2v_args.extend([
+            '-i', 'vmx',
+            '-it', 'ssh',
             ])
 
     if 'rhv_url' in data:
@@ -336,10 +381,13 @@ def wrapper(data, state_file, v2v_log):
         env['LIBGUESTFS_BACKEND'] = 'direct'
     if 'virtio_win' in data:
         env['VIRTIO_WIN'] = data['virtio_win']
+    if agent_sock is not None:
+        env['SSH_AUTH_SOCK'] = agent_sock
 
     proc = None
     with open(v2v_log, 'w') as log:
-        logging.info('Starting virt-v2v as: %r', v2v_args)
+        logging.info('Starting virt-v2v as: %r, environment: %r',
+                     v2v_args, env)
         proc = subprocess.Popen(
                 v2v_args,
                 stderr=subprocess.STDOUT,
@@ -394,8 +442,115 @@ def write_password(password, password_files):
     return pfile[1]
 
 
+def spawn_ssh_agent():
+    try:
+        out = subprocess.check_output(['ssh-agent'])
+        logging.debug('ssh-agent: %s' % out)
+        sock = re.search(br'^SSH_AUTH_SOCK=([^;]+);', out, re.MULTILINE)
+        pid = re.search(br'^echo Agent pid ([0-9]+);', out, re.MULTILINE)
+        if not sock or not pid:
+            logging.error(
+                'Incomplete match of ssh-agent output; sock=%r; pid=%r',
+                sock, pid)
+            return None, None
+        agent_sock = sock.group(1).decode()
+        agent_pid = int(pid.group(1))
+        logging.info('SSH Agent started with PID %d', agent_pid)
+    except subprocess.CalledProcessError:
+        logging.error('Failed to start ssh-agent')
+        return None, None
+    env = os.environ.copy()
+    env['SSH_AUTH_SOCK'] = agent_sock
+    ret_code = subprocess.call(['ssh-add'], env=env)
+    if ret_code != 0:
+        logging.error('Failed to add SSH keys to the agent! ssh-add'
+                      ' terminated with return code %d', ret_code)
+        os.kill(agent_pid, signal.SIGTERM)
+        return None, None
+    return agent_pid, agent_sock
+
+
+def check_install_drivers(data):
+    if 'virtio_win' in data and os.path.isabs(data['virtio_win']):
+        full_path = data['virtio_win']
+    else:
+        iso_domain = find_iso_domain()
+
+        iso_name = data.get('virtio_win')
+        if iso_name is not None:
+            if iso_domain is None:
+                error('ISO domain not found')
+        else:
+            if iso_domain is None:
+                # This is not an error
+                logging.warning('ISO domain not found' +
+                                ' (but install_drivers is true).')
+                data['install_drivers'] = False
+                return
+
+            # (priority, pattern)
+            patterns = [
+                (4, br'RHV-toolsSetup_([0-9._]+)\.iso'),
+                (3, br'RHEV-toolsSetup_([0-9._]+)\.iso'),
+                (2, br'oVirt-toolsSetup_([a-z0-9._-]+)\.iso'),
+                (1, br'virtio-win-([0-9.]+).iso'),
+                ]
+            patterns = [(p[0], re.compile(p[1], re.IGNORECASE))
+                        for p in patterns]
+            best_name = None
+            best_version = None
+            best_priority = -1
+            for fname in os.listdir(iso_domain):
+                if not os.path.isfile(os.path.join(iso_domain, fname)):
+                    continue
+                for priority, pat in patterns:
+                    m = pat.match(fname)
+                    if not m:
+                        continue
+                    version = m.group(1)
+                    logging.debug('Matched ISO %r (priority %d)',
+                                  fname, priority)
+                    if best_version is None or \
+                            (best_version < version and
+                             best_priority <= priority):
+                        best_name = fname
+                        best_version = version
+
+            if best_name is None:
+                # Nothing found, this is not an error
+                logging.warn('Could not find any ISO with drivers' +
+                             ' (but install_drivers is true).')
+                data['install_drivers'] = False
+                return
+            iso_name = best_name
+
+        full_path = os.path.join(iso_domain, iso_name)
+
+    if not os.path.isfile(full_path):
+        error("'virtio_win' must be a path or file name of image in "
+              "ISO domain")
+    data['virtio_win'] = full_path
+    logging.info("virtio_win (re)defined as: %s", data['virtio_win'])
+
+
 ###########
 
+# Read and parse input -- hopefully this should be safe to do as root
+data = json.load(sys.stdin)
+
+# NOTE: this is just pre-check to find out whether we can run as vdsm user at
+# all. This is not validation of the input data!
+if 'export_domain' in data:
+    # Need to be root to mount NFS share
+    VDSM = False
+    # Cannot use libvirt backend as root on VDSM host due to permissions
+    DIRECT_BACKEND = True
+
+if VDSM:
+    make_vdsm(data)
+
+# The logging is delayed after we now which user runs the wrapper. Otherwise we
+# would have two logs.
 log_tag = '%s-%d' % (time.strftime('%Y%m%dT%H%M%S'), os.getpid())
 v2v_log = os.path.join(VDSM_LOG_DIR, 'v2v-import-%s.log' % log_tag)
 wrapper_log = os.path.join(VDSM_LOG_DIR, 'v2v-import-%s-wrapper.log' % log_tag)
@@ -405,10 +560,8 @@ logging.basicConfig(
     level=LOG_LEVEL,
     filename=wrapper_log,
     format='%(asctime)s:%(levelname)s: %(message)s (%(module)s:%(lineno)d)')
-logging.info('Wrapper version %d', VERSION)
 
-if VDSM:
-    make_vdsm()
+logging.info('Wrapper version %d, uid=%d', VERSION, os.getuid())
 
 logging.info('Will store virt-v2v log in: %s', v2v_log)
 logging.info('Will store state file in: %s', state_file)
@@ -416,19 +569,10 @@ logging.info('Will store state file in: %s', state_file)
 password_files = []
 
 try:
-    logging.info('Processing input data')
-    data = json.load(sys.stdin)
-
     # Make sure all the needed keys are in data. This is rather poor
     # validation, but...
-    for k in [
-            'vm_name',
-            'vmware_fingerprint',
-            'vmware_uri',
-            'vmware_password',
-            ]:
-        if k not in data:
-            error('Missing argument: %s' % k)
+    if 'vm_name' not in data:
+            error('Missing vm_name')
 
     # Output file format (raw or qcow2)
     if 'output_format' in data:
@@ -441,9 +585,18 @@ try:
     # Transports (only VDDK for now)
     if 'transport_method' not in data:
         error('No transport method specified')
-    if data['transport_method'] != 'vddk':
+    if data['transport_method'] not in ('ssh', 'vddk'):
         error('Unknown transport method: %s', data['transport_method'])
 
+    if data['transport_method'] == 'vddk':
+        for k in [
+                'vmware_fingerprint',
+                'vmware_uri',
+                'vmware_password',
+                ]:
+            if k not in data:
+                error('Missing argument: %s' % k)
+
     # Targets (only export domain for now)
     if 'rhv_url' in data:
         for k in [
@@ -473,18 +626,12 @@ try:
 
     # Virtio drivers
     if 'virtio_win' in data:
-        if not os.path.isabs(data['virtio_win']):
-            iso_domain = find_iso_domain()
-            if iso_domain is None:
-                error('ISO domain not found')
-            full_path = os.path.join(iso_domain, data['virtio_win'])
-        else:
-            full_path = data['virtio_win']
-        if not os.path.isfile(full_path):
-            error("'virtio_win' must be a path or file name of image in "
-                  "ISO domain")
-        data['virtio_win'] = full_path
-        logging.info("virtio_win (re)defined as: %s", data['virtio_win'])
+        # This is for backward compatibility
+        data['install_drivers'] = True
+    if 'install_drivers' in data:
+        check_install_drivers(data)
+    else:
+        data['install_drivers'] = False
 
     # Allocation type
     if 'allocation' in data:
@@ -492,17 +639,18 @@ try:
             error('Invalid value for allocation type: %r' % data['allocation'])
     else:
         # Check storage domain type and decide on suitable allocation type
-        # Note: This is only temporary. We should get the info from the caller in
-        # the future.
+        # Note: This is only temporary. We should get the info from the caller
+        # in the future.
         domain_type = None
         with sdk_connection(data) as c:
             service = c.system_service().storage_domains_service()
-            domains = service.list(search='name="%s"' % str(data['rhv_storage']))
+            domains = service.list(search='name="%s"' %
+                                   str(data['rhv_storage']))
             if len(domains) != 1:
                 error('Found %d domains matching "%s"!' % data['rhv_storage'])
             domain_type = domains[0].storage.type
         logging.info('Storage domain "%s" is of type %r', data['rhv_storage'],
-                    domain_type)
+                     domain_type)
         data['allocation'] = 'sparse'
         if domain_type in PREALLOCATED_STORAGE_TYPES:
             data['allocation'] = 'preallocated'
@@ -514,12 +662,14 @@ try:
 
     # Store password(s)
     logging.info('Writing password file(s)')
-    data['vmware_password_file'] = write_password(data['vmware_password'],
-                                                  password_files)
+    if 'vmware_password' in data:
+        data['vmware_password_file'] = write_password(data['vmware_password'],
+                                                      password_files)
     if 'rhv_password' in data:
         data['rhv_password_file'] = write_password(data['rhv_password'],
                                                    password_files)
 
+    # TODO: create (and manage) state file before dumping the json
     # Send some useful info on stdout in JSON
     print(json.dumps({
         'v2v_log': v2v_log,
@@ -530,7 +680,15 @@ try:
     # Let's get to work
     logging.info('Daemonizing')
     daemonize()
-    wrapper(data, state_file, v2v_log)
+    agent_pid = None
+    agent_sock = None
+    if data['transport_method'] == 'ssh':
+        agent_pid, agent_sock = spawn_ssh_agent()
+        if agent_pid is None:
+            raise RuntimeError('Failed to start ssh-agent')
+    wrapper(data, state_file, v2v_log, agent_sock)
+    if agent_pid is not None:
+        os.kill(agent_pid, signal.SIGTERM)
 
     # Remove password files
     logging.info('Removing password files')
